/************************************************************************
 * FILE: nutils.c							*
 *	 Numerical utility functions for the wftx system 		*
 *									*
 * Author: Peter Rajcani, March 1995.					*
 ************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <math.h>

#define LEFT	1
#define RIGHT	2
#define TRUE	1
#define FALSE	0

typedef float   imgtype;


/*
 * Allocate space for a vector
 */

float *alloc_vect (size)
int	size;
{
   float  *result;

   if ((result = (float *) calloc (size, sizeof(float))) == NULL) {
		printf ("Memory allocation error\n"); exit(1); }

   return (result);
}

/*
 * Integer version of the above function
 */

int *alloc_int_array (size)
int	size;
{
   int  *result;

   if ((result = (int *) calloc (size, sizeof(int))) == NULL) {
		printf ("allox_vector : Memory allocation error\n"); exit(1); }

   return (result);
}


/*
 * Allocate m x n matrix
 */

float **alloc_matrix (rows, columns)
int 	rows, columns;
{
  float  **matrix;
  int	 i;
 
  
    if ((matrix = (float **) calloc (rows, sizeof (float*))) == NULL) {
	printf ("ERROR: alloc_matrix: Memory allocation error\n"); exit (-1);
    }

    for (i = 0; i < rows; i++) 
        matrix[i] = alloc_vect(columns);

    return (matrix);
}


/*
 * Allocate square matrix
 */

float **alloc_sq_matrix (size)
int size;
{
  float **matrix;

  matrix = alloc_matrix (size, size);
 
  return (matrix);
}

/*
 * free memory space occupied by a matrix
 */

void free_matrix (matrix, nrows)
float **matrix;
int   nrows;
{
  int   i;

  for (i = 0; i < nrows; i++)
	free (matrix[i]);

  free(matrix);
}

/*
 * Compute inner product of two vectors
 */

float	inner_prod (a, b, n)
float   *a, *b;
int	n;
{
  float	 sum;
  int    i;
 
  for (i = 0, sum = 0; i < n; i++)
	sum += a[i]*b[i];

  return (sum);
}


/*
 * Compute norm of a vector
 */

float	compute_norm (vector, size)
float   *vector;
int	size;
{
  return (inner_prod (vector, vector, size));
}

  
/*
 * Copy vector a into vector b
 */

void 	copy_vect (a, b, n)
float	*a, *b;
int	n;
{
  int	i;

  for (i = 0; i < n; i++) b[i] = a[i];

}


/*
 * Orthogonalization procedure
 */

 void orthogonalize (ortho_vector, ortho_norm, new_vector, remainder,
		     size, nbase)
 float	**ortho_vector, *ortho_norm, *new_vector, *remainder;
 int	size, nbase;
 {
    int    i, j ;
    float  c;

    copy_vect (new_vector, remainder, size);
    for (i = 0; i < nbase; i++) {
	c = inner_prod (ortho_vector[i], new_vector, size)/ortho_norm[i];
	for (j = 0; j < size; j++)
	    remainder[j] -= c * ortho_vector[i][j];
    }
 }


/*
 * Multiply a square matrix with a vector
 */

void vec_matrix_mult (vector, matrix, result, size, type)
float *vector, **matrix, *result; 
int   size, type;
{
  int  i, j;
  float res;

   for (i = 0; i < size; i++) {
        for (res = 0, j = 0; j < size; j++) {
  	     switch (type) {
                     case LEFT : 
			     res += vector[j] * matrix[j][i];
			     break;
		     case RIGHT:
			     res += vector[j] * matrix[i][j];
			     break;
	     }
         }
         result[i] = res;
  }
}


/******************************************************************** 
 * The following two functions are taken from "Numerical Recipes in C".
 * For details and description see the book pp. 44-48.
 ********************************************************************/

/*********************************************************
 * LU Decomposition
 * 
 * Parameters: 	a - regular n x n matrix
 *		n - number of rows (coulumns) of the matrix a 
 *	       	indx - array storing the row permutation
 *		d - records if number of interchanges was
 *		    even or odd.	
 **********************************************************/


int ludcmp(a,n,indx,d)
int n,*indx;
float **a,*d;
{
	int i,imax,j,k;
	float big,dum,sum,temp;
	float *vv;

	vv=alloc_vect(n+1);
	*d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) {
		    printf("ERROR: Singular matrix in routine LUDCMP\n");
		    return (FALSE);
		}
		vv[i]=1.0/big;
	}
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i][j];
			for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i][j];
			for (k=1;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) {
		    printf("ERROR: Singular matrix in routine LUDCMP\n");
		    return (FALSE);
		}
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n;i++) a[i][j] *= dum;
		}
	}
	free(vv);
        return (TRUE);
}


/*****************************************************************
 * Routine for forward substitution and backsubstitution
 * which solves a linear system of equations after LU decomposition.
 *
 * Parameters : a - n x n matrix after LU decomposition.
 *		n - number of rows (columns) in the matrix a
 *		indx - integer array that stores the row permutation.
 *			This array is created by the ludcmp routine.
 *		b - the RHS of the linear system
 ********************************************************************/

void lubksb(a,n,indx,b)
float **a,b[];
int n,*indx;
{
	int i,ii=0,ip,j;
	float sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}

/********* print out vector & matrix of type float *****************/

void print_vector (outfile, vector, size)
FILE	*outfile;
float	*vector;
int	size;
{
  int 	i;

  for (i=0; i<size; i++) {
      if (outfile != NULL) 
          fprintf (outfile,"index: %d, value: %5.3f \n", i, vector[i]);
      else
          printf ("index: %d, value: %5.3f \n", i, vector[i]);
  }

}

void print_matrix (outfile, matrix, rows, columns)
FILE	*outfile;
float	**matrix;
int	rows, columns;
{
  int	i, j;

  if (outfile != NULL) {
      for (i=0; i < rows; i++) {
          fprintf (outfile, "row %d: ", i);
          for (j=0; j < columns; j++) 
	      fprintf (outfile, "%5.3f  ", matrix[i][j]);
          fprintf (outfile, "\n");
      }
   } else {
      for (i=0; i < rows; i++) {
          printf ("row %d: ", i);
          for (j=0; j < columns; j++) 
	      printf ("%5.3f  ", matrix[i][j]);
      printf ("\n");
      }
  }
}
	
/*********************************************************
 *  Image utilities
 ********************************************************/

imgtype **alloc_image (rows, columns)
int	rows, columns;
{
  int	i;
  imgtype **raw_image;

  if ((raw_image = (imgtype **) calloc (rows, 
                    sizeof (imgtype *))) == NULL) {
	            printf ("Error in allocating image array\n");
	            exit (1); }  

   for (i = 0; i < rows; i++) {
     	if ((raw_image [i] = (imgtype *) calloc (columns, 
                            sizeof (imgtype))) == NULL) {
	       	printf ("Error in allocating image array\n"); exit (1); }  
   }

   return (raw_image);
}

void free_image (raw_image, rows)
imgtype **raw_image;
int	      rows;
{
  int	i;

  for (i=0; i<rows; i++)
       free (raw_image[i]);

  free (raw_image);
}

void copy_image (src_img, dest_img, rows, columns)
imgtype	**src_img, **dest_img;
int		rows, columns;
{
   int 	i, j;

   for (i=0; i<rows; i++) 
	for (j=0; j<columns; j++) 
	     dest_img[i][j] = src_img[i][j];
}
